defmodule ClickHouse.Interface.HTTP.Client do
  @moduledoc false

  alias ClickHouse.ConnectionError

  @opts_schema %{
    checkout_timeout: [type: :timeout, required: false, default: 8_000],
    connect_timeout: [type: :timeout, required: false, default: 8_000],
    recv_timeout: [type: :timeout, required: false, default: 15_000],
    buffer_size: [type: :integer, required: false],
    wait_end_of_query: [type: :integer, required: false],
    session_id: [type: :binary, required: false],
    session_timeout: [type: :integer, required: false],
    database: [type: :binary, required: false],
    settings: [type: :map, required: false],
    content_encoding: [type: :atom, inclusion: [:gzip, :deflate], required: false],
    accept_encoding: [type: :atom, inclusion: [:gzip, :deflate], required: false],
    decompress: [type: :boolean, required: false]
  }

  @closed_retries 3
  @accept_header "Accept-Encoding"
  @content_header "Content-Encoding"
  @format_header "X-ClickHouse-Format"
  @error_header "X-ClickHouse-Exception-Code"
  @error_regex ~r/(?<=displayText\(\) = )(.*?)(?=: )/
  @error_types %{
    "Coordination::Exception" => ClickHouse.CoordinationError,
    "DB::Exception" => ClickHouse.DatabaseError,
    "DB::NetException" => ClickHouse.NetworkError,
    "DB::ParsingException" => ClickHouse.ParsingError
  }

  ################################
  # Public API
  ################################

  @doc false
  def request(pool, [url], body, opts) do
    request(pool, url, body, opts)
  end

  def request(pool, urls, body, opts) when is_list(urls) do
    url = Enum.random(urls)
    request(pool, url, body, opts)
  end

  def request(pool, url, body, opts) do
    opts = KeywordValidator.validate!(opts, @opts_schema)
    url = build_url(url, opts)
    headers = build_headers(opts)
    body = maybe_compress(body, opts)
    http_opts = build_http_opts(pool, opts)
    do_request(1, url, headers, body, http_opts, opts)
  end

  ################################
  # Private API
  ################################

  defp do_request(attempt, url, headers, body, http_opts, opts) do
    case :hackney.request(:post, url, headers, body, http_opts) do
      {:ok, 200, headers, body} ->
        build_result(headers, body, opts)

      {:error, :closed} when attempt < @closed_retries ->
        do_request(attempt + 1, url, headers, body, http_opts, opts)

      error ->
        build_error(error)
    end
  end

  defp build_url(url, opts) do
    settings = Keyword.get(opts, :settings, %{})

    params =
      opts
      |> maybe_compression_param()
      |> Keyword.take([
        :buffer_size,
        :wait_end_of_query,
        :session_id,
        :session_timeout,
        :database,
        :enable_http_compression
      ])
      |> Enum.into(%{})
      |> Map.merge(settings)

    case URI.encode_query(params) do
      "" -> url
      params -> "#{url}?#{params}"
    end
  end

  defp maybe_compression_param(opts) do
    if Keyword.has_key?(opts, :accept_encoding) do
      Keyword.put(opts, :enable_http_compression, 1)
    else
      opts
    end
  end

  defp build_headers(opts) do
    []
    |> maybe_accept_encoding(opts)
    |> maybe_content_encoding(opts)
  end

  defp maybe_accept_encoding(headers, opts) do
    if accept_encoding = Keyword.get(opts, :accept_encoding) do
      [{@accept_header, to_string(accept_encoding)} | headers]
    else
      headers
    end
  end

  defp maybe_content_encoding(headers, opts) do
    if content_encoding = Keyword.get(opts, :content_encoding) do
      [{@content_header, to_string(content_encoding)} | headers]
    else
      headers
    end
  end

  defp maybe_compress(statement, opts) do
    if compression = Keyword.get(opts, :content_encoding) do
      compress(compression, statement)
    else
      statement
    end
  end

  defp compress(:gzip, statement), do: :zlib.gzip(statement)
  defp compress(:deflate, statement), do: :zlib.compress(statement)

  defp build_http_opts(pool, opts) do
    opts
    |> Keyword.take([:checkout_timeout, :connect_timeout, :recv_timeout])
    |> Keyword.merge(pool: pool, with_body: true)
  end

  defp build_result(headers, body, opts) do
    format = find_header(headers, @format_header)
    encoding = find_header(headers, @content_header)
    {body, compressed} = maybe_decompress(encoding, body, opts)
    meta = get_meta(headers)
    {:ok, {body, format, meta, compressed}}
  end

  defp build_error({:error, reason}) do
    {:error, %ConnectionError{message: to_string(reason)}}
  end

  defp build_error({:ok, _code, headers, body}) do
    encoding = find_header(headers, @content_header)
    {body, _} = maybe_decompress(encoding, body, decompress: true)
    type = get_error_type(body)
    code = find_header(headers, @error_header)
    message = get_error_message(body)
    meta = get_meta(headers)
    error = struct(type, code: code, message: message, meta: meta)

    {:error, error}
  end

  defp get_meta(headers) do
    %{
      query_id: find_header(headers, "X-ClickHouse-Query-Id"),
      server_display_name: find_header(headers, "X-ClickHouse-Server-Display-Name"),
      timezone: find_header(headers, "X-ClickHouse-Timezone"),
      summary: find_header(headers, "X-ClickHouse-Summary")
    }
  end

  defp get_error_type(body) do
    case Regex.run(@error_regex, body, capture: :first) do
      [error] -> Map.fetch!(@error_types, error)
    end
  end

  defp get_error_message(body) do
    body
    |> String.split("e.displayText() = ")
    |> List.last()
  end

  defp find_header([], _), do: nil

  defp find_header([{key, val} | _headers], header) when key == header do
    val
  end

  defp find_header([_ | headers], header) do
    find_header(headers, header)
  end

  defp maybe_decompress(nil, body, _opts), do: {body, false}

  defp maybe_decompress(compression, body, opts) do
    if Keyword.get(opts, :decompress) do
      {decompress(compression, body), false}
    else
      {body, true}
    end
  end

  defp decompress("gzip", body), do: :zlib.gunzip(body)
  defp decompress("deflate", body), do: :zlib.uncompress(body)
end
